#! /usr/bin/env python3
#
# Copyright (C) 2024 MicroSys Electronics GmbH
# Author: Kay Potthoff <kay.potthoff@microsys.de>
#

import sys
import getopt
import os

__version__ = "1.0.0"

HASH_ALGO = "sha256"

class DtsEntry:

    def __init__(self, name, indent=0):
        self.name = name
        self.indent = int(indent)
        self.indent_step = 4
        self.indent_stack = []
        self.parent = None

    def getName(self):
        return self.name

    def setName(self, name):
        self.name = name

    def getParent(self):
        return self.parent

    def setParent(self, parent):
        self.parent = parent

    def pushIndent(self, next_indent=None):
        self.indent_stack.append(self.getIndent())
        if next_indent:
            self.setIndent(next_indent)

    def popIndent(self):
        if len(self.indent_stack) > 0:
            self.setIndent(self.indent_stack.pop())

    def setIndent(self, indent):
        if indent >= 0:
            self.indent = int(indent)
        else:
            self.indent = 0

    def getIndent(self):
        return self.indent

    def resetIndent(self):
        self.setIndent(0)

    def incIndent(self, n=1):
        self.indent += n

    def decIndent(self, n=1):
        self.indent -= n

    def tab(self):
        return " " * (self.indent*self.indent_step)

    def write(self, fh):
        raise NotImplementedError()

    def isNode(self):
        return False

class DtsProperty(DtsEntry):

    def __init__(self, name, value=None):
        super().__init__(name)
        self.value = value

    def __str__(self):
        return str(self.getValue())

    def getValue(self):
        return self.value

    def setValue(self, value):
        self.value = value

    def emit(self):
        return self.tab() + self.getName() + " = " + str(self)

    def write(self, fh):
        print(self.emit()+";", file=fh)

class DtsStringProperty(DtsProperty):

    def __init__(self, name, value=None):
        if not value:
            value = ''
        super().__init__(name, value)

    def __str__(self):
        return '"' + str(self.getValue()) + '"'

class DtsBooleanProperty(DtsProperty):

    def __init__(self, name):
        super().__init__(name, True)

    def emit(self):
        return self.tab() + self.getName()

class DtsIncBinProperty(DtsProperty):

    def __init__(self, name, value=None):
        if not value:
            value = ''
        super().__init__(name, value)

    def __str__(self):
        return '/incbin/("' + str(self.getValue()) + '")'

class DtsIntBaseProperty(DtsProperty):

    def __init__(self, name, value=None):
        super().__init__(name, value)
        self.base = 0

    def getBase(self):
        return self.base

    def setBase(self, base):
        self.base = base

    def getFormatStr(self):
        if (self.base == 0) or (self.base == 10):
            return "{0:d}"
        elif (self.base == 16):
            return "0x{0:x}"
        else:
            return "{0}"

class DtsIntProperty(DtsIntBaseProperty):

    def __init__(self, name, value=None):
        if value is None:
            value = -1
        if isinstance(value, str):
            value = int(value, 0)
        super().__init__(name, value)
        if (name == "offset") or (name == "reg") \
            or (name == "entry") or (name == "load"):
            self.setBase(16)
        else:
            self.setBase(10)

    def __str__(self):
        fmt = self.getFormatStr()
        return "<" + fmt.format(self.getValue()) + ">"

class DtsIntListProperty(DtsIntBaseProperty):

    def __init__(self, name, value=None):
        if not value:
            value = []
        if not isinstance(value, list):
            value = list(value)
        vl = []
        for v in value:
            if isinstance(v, str):
                v = int(v, 0)
            vl.append(v)
        super().__init__(name, vl)
        self.setBase(16)

    def __str__(self):
        fmt = self.getFormatStr()
        a = [fmt.format(v) for v in self.getValue()]
        return "<" + " ".join(a) + ">"

class DtsRegProperty(DtsIntListProperty):

    def __init__(self, address, size):
        super().__init__("reg")
        self.setBase(16)
        self.setAddress(address)
        self.setSize(size)

    def setAddress(self, address):
        self.address = address

    def getAddress(self):
        return self.address

    def setSize(self, size):
        self.size = size

    def getSize(self):
        return self.size

    def getValue(self):
        adr_w = 1
        sz_w = 1
        parent = self.getParent()
        if parent:
            w = parent.getAddressCells()
            if w > 0:
                adr_w = w
            w = parent.getSizeCells()
            if w > 0:
                sz_w = w
        vl = []
        for shift in reversed(range(adr_w)):
            a = (self.address >> (shift*32)) & 0xffffffff
            vl.append(a)
        for shift in reversed(range(sz_w)):
            a = (self.size >> (shift*32)) & 0xffffffff
            vl.append(a)
        return vl

class DtsNode(DtsEntry):

    def __init__(self, name, address_cells=-1, size_cells=-1):
        super().__init__(name)
        self.child = {}
        self.setAddressCells(address_cells)
        self.setSizeCells(size_cells)

    def isNode(self):
        return True

    def setAddressCells(self, address_cells):
        self.address_cells = address_cells
        if address_cells > 0:
            self["#address-cells"] = address_cells

    def getAddressCells(self):
        if self.address_cells > 0:
            return self.address_cells
        parent = self.getParent()
        if parent:
            return parent.getAddressCells()
        return 1

    def setSizeCells(self, size_cells):
        self.size_cells = size_cells
        if size_cells > 0:
            self["#size-cells"] = size_cells

    def getSizeCells(self):
        if self.size_cells > 0:
            return self.size_cells
        parent = self.getParent()
        if parent:
            return parent.getSizeCells()
        return 1

    def set(self, entry):
        self.child[entry.getName()] = entry
        entry.setParent(self)

    def get(self, name):
        return self.child[name]

    def __contains__(self, item):
        return item in self.child

    def __getitem__(self, name):
        return self.get(name)

    def __setitem__(self, name, value):
        if isinstance(value, str):
            self.set(DtsStringProperty(name, value))
        elif isinstance(value, bool):
            self.set(DtsBooleanProperty(name))
        elif isinstance(value, int):
            self.set(DtsIntProperty(name, value))
        elif isinstance(value, list):
            self.set(DtsIntListProperty(name, value))
        elif isinstance(value, DtsProperty):
            self.set(value)
        elif isinstance(value, DtsEntry):
            if name in self.child:
                self.child[name].set(value)

    def incIndent(self, n=1):
        super().incIndent(n)
        for c in self.child:
            self.child[c].incIndent()

    def decIndent(self, n=1):
        super().decIndent(n)
        for c in self.child:
            self.child[c].decIndent()

    def write(self, fh):
        print("", file=fh)
        tab = self.tab()
        print(tab+self.getName()+ " {", file=fh)

        # Note: Properties must precede subnodes

        # 1. all properties:
        for c in self.child:
            if isinstance(self.child[c], DtsProperty):
                self.child[c].incIndent()
                self.child[c].write(fh)
                self.child[c].decIndent()

        # 2. all subnodes:
        for c in self.child:
            if isinstance(self.child[c], DtsNode):
                self.child[c].incIndent()
                self.child[c].write(fh)
                self.child[c].decIndent()

        print(tab+"};", file=fh)

    def getSubnodes(self):
        nodes = []
        for c in self.child:
            if isinstance(self.child[c], DtsNode):
                nodes.append(self.child[c])
        return nodes

    def parseParameter(self, param):
        args = param.split(':')
        image = args[0]
        args = args[1:]
        arg = {}
        for a in args:
            p = a.split('=')
            if len(p) > 1:
                arg[p[0]] = p[1]
        return (image, arg)

class DtsFITNode(DtsNode):

    def __init__(self, name):
        super().__init__(name)

        for n in ["images", "configurations"]:
            self.set(DtsNode(n))

    def addImage(self, name, description=None, hash=HASH_ALGO):
        self["images"] = DtsNode(name)
        if description:
            self["images"][name]["description"] = description
        self["images"][name]["arch"] = "arm64"

        if hash:
            self["images"][name] = DtsNode("hash-1")
            self["images"][name]["hash-1"]["algo"] = hash

        # Note:
        #'compression' nodes for ramdisks are deprecated
        if not name.startswith("ramdisk"):
            self["images"][name]["compression"] = "none"

        if name.startswith("kernel"):
            self["images"][name]["type"] = "kernel"
            self["images"][name]["os"] = "linux"

        if name.startswith("fdt"):
            self["images"][name]["type"] = "flat_dt"

        if name.startswith("ramdisk"):
            self["images"][name]["type"] = "ramdisk"
            self["images"][name]["os"] = "linux"

    def setImageEntry(self, name, address):
        self["images"][name].set(DtsIntProperty("entry", address))
        if name.startswith("kernel"):
            self["images"][name].set(DtsIntProperty("load", address))

    def setImage(self, name, image, hash=HASH_ALGO):
        if not name in self["images"]:
            self.addImage(name, hash=hash)
        (image, arg) = self.parseParameter(image)
        if os.path.isfile(image):
            img = os.path.realpath(image)
            self["images"][name].set(DtsIncBinProperty("data", img))
        else:
            self["images"][name]["data"] = image
        # Note:
        #'compression' nodes for ramdisks are deprecated
        if image.endswith(".gz") and (not name.startswith("ramdisk")):
            self["images"][name]["compression"] = "gzip"
        if "entry" in arg:
            self.setImageEntry(name, arg["entry"])
        if "load" in arg:
            self["images"][name].set(DtsIntProperty("load", arg["load"]))

    def addConfiguration(self, name, description=None, hash=HASH_ALGO):
        self["configurations"] = DtsNode(name)
        if description:
            self["configurations"][name]["description"] = description
        else:
            self["configurations"][name]["description"] = f"configuration {name}"

        if hash:
            self["configurations"][name] = DtsNode("hash-1")
            self["configurations"][name]["hash-1"]["algo"] = hash

    def setDefaultConfiguration(self, config):
        self["configurations"]["default"] = config

    def finalizeConfigs(self, hash=HASH_ALGO):
        subnode = self["images"].getSubnodes()
        kernel = []
        fdt = []
        ramdisk = []
        for sn in subnode:
            if sn.getName().startswith("kernel"):
                kernel.append(sn)
            elif sn.getName().startswith("fdt"):
                fdt.append(sn)
            elif sn.getName().startswith("ramdisk"):
                ramdisk.append(sn)
        cfg_no = 0
        if len(ramdisk) > 0:
            for k in kernel:
                for f in fdt:
                    for r in ramdisk:
                        cfg = "ramboot-{0}".format(cfg_no)
                        self.addConfiguration(cfg, hash=hash)
                        if cfg_no == 0:
                            self.setDefaultConfiguration(cfg)
                        self["configurations"][cfg]["kernel"] = k.getName()
                        self["configurations"][cfg]["fdt"] = f.getName()
                        self["configurations"][cfg]["ramdisk"] = r.getName()
                        cfg_no = cfg_no + 1
        else:
            for k in kernel:
                for f in fdt:
                    cfg = "conf-{0}".format(cfg_no)
                    self.addConfiguration(cfg, hash=hash)
                    if cfg_no == 0:
                        self.setDefaultConfiguration(cfg)
                    self["configurations"][cfg]["kernel"] = k.getName()
                    self["configurations"][cfg]["fdt"] = f.getName()
                    cfg_no = cfg_no + 1

class ITS:

    VERSION = "/dts-v1/"

    def __init__(self, root_node, description=None):
        self.setRoot(root_node)
        if description:
            self.root["description"] = description

    def setDescription(self, description):
        self.root["description"] = description

    def setRoot(self, root_node):
        self.root = root_node

    def getRoot(self):
        return self.root

    def set(self, entry):
        self.getRoot().set(entry)

    def get(self, name):
        return self.getRoot().get(name)

    def __getitem__(self, name):
        return self.get(name)

    def __setitem__(self, name, value):
        self.getRoot()[name] = value

    def write(self, fh):
        print("/* generated file; do not edit */", file=fh)
        print(self.VERSION+";", file=fh)
        self.getRoot().write(fh)

class FITImage(ITS):

    def __init__(self, description=None):
        super().__init__(DtsFITNode("/"), description)

    def addImage(self, name, description=None, hash=HASH_ALGO):
        self.getRoot().addImage(name, description, hash=hash)

    def setImage(self, name, image, hash=HASH_ALGO):
        self.getRoot().setImage(name, image, hash=hash)

    def setImageEntry(self, name, address):
        self.getRoot().setImageEntry(name, address)

    def addConfiguration(self, name, description=None, hash=HASH_ALGO):
        self.getRoot().addConfiguration(name, description, hash=hash)

    def setDefaultConfiguration(self, config):
        self.getRoot().setDefaultConfiguration(config)

    def finalizeConfigs(self, hash=HASH_ALGO):
        self.getRoot().finalizeConfigs(hash=hash)

class RamdiskITS(FITImage):

    def __init__(self):
        super().__init__("U-Boot ITS for ramdisk")

class Binman(ITS):

    def __init__(self, description=None):
        super().__init__(DtsNode("/"), description)
        self.getRoot().set(DtsNode("binman"))
        self["binman"]["sort-by-offset"] = True

    def setFilename(self, fname):
        self["binman"]["filename"] = fname

    def setPadByte(self, pad_byte):
        pad = DtsIntProperty("pad-byte", pad_byte)
        pad.setBase(16)
        self["binman"].set(pad)

    def setEType(self, etype):
        (image, arg) = self.getRoot().parseParameter(etype)

        name = "binary"
        offset = 0

        if "name" in arg:
            name = arg["name"]

        if "offset" in arg:
            offset = int(arg["offset"], 0)

        entry = DtsNode(name)
        entry["filename"] = image
        entry["offset"] = offset

        if name == "null":
            entry["size"] = 0

        self["binman"].set(entry)

def usage():
    prog = os.path.basename(sys.argv[0])
    print("{0} {1}\n".format(prog, __version__))
    print("usage: {0} [options] ...".format(prog))
    print("\nOptions:")
    print("  -h|--help                  print this help")

def main():

    filename = None
    kernel = []
    fdt = []
    ramdisk = []
    etype = []
    binman = None
    hash = HASH_ALGO

    try:
        opts, argv = getopt.getopt(sys.argv[1:],
                                   "hf:",
                                   ["help",
                                    "no-hash",
                                    "file=",
                                    "binman=",
                                    "kernel=",
                                    "fdt=",
                                    "ramdisk=",
                                    "etype=",
                                    "hash="])
    except getopt.GetoptError as err:
        print("Error: {0}".format(err), file=sys.stderr)
        usage()
        sys.exit(1)

    for option, optarg in opts:
        if option in ("-h", "--help"):
            usage()
            sys.exit(0)
        elif option in ("-f", "--file"):
            filename = optarg
        elif option in ("--binman"):
            binman = optarg
        elif option in ("--kernel"):
            kernel.append(optarg)
        elif option in ("--fdt"):
            fdt.append(optarg)
        elif option in ("--ramdisk"):
            ramdisk.append(optarg)
        elif option in ("--etype"):
            etype.append(optarg)
        elif option in ("--hash"):
            hash = optarg
        elif option in ("--no-hash"):
            hash = None
        else:
            print("Error: unknown option \"{0}\"!".format(option), file=sys.stderr)
            sys.exit(1)

    if binman:

        its = Binman()

        its.setFilename(binman)
        its.setPadByte(0xff)

        for i in range(len(etype)):
            its.setEType(etype[i])

    else:

        its = RamdiskITS()

        for i in range(len(kernel)):
            n = "kernel-{0}".format(i)
            its.addImage(n, "Linux kernel", hash=hash)
            its.setImage(n, kernel[i], hash=hash)

        for i in range(len(fdt)):
            n = "fdt-{0}".format(i)
            its.addImage(n, "Flattened Device Tree blob", hash=hash)
            its.setImage(n, fdt[i], hash=hash)

        for i in range(len(ramdisk)):
            n = "ramdisk-{0}".format(i)
            its.addImage(n, "Ramdisk image", hash=hash)
            its.setImage(n, ramdisk[i], hash=hash)

        if len(ramdisk) == 0:
            its.setDescription("U-Boot ITS Linux Kernel Image")

        its.finalizeConfigs(hash=hash)

    if filename:
        with open(filename, "w") as fh:
            its.write(fh)
    else:
        its.write(sys.stdout)

    return 0

if __name__ == "__main__":
    sys.exit(main())

#######################################
# Local Variables:
# mode: Python
# tab-width: 4
# End:
# kate: space-indent on; indent-width 4; mixedindent off; indent-mode python;
# vi: set tabstop=4 shiftwidth=4:
# vim: set filetype=python:
